/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include "lib.h"
#include "iphash.h"
#define CLASS_SIZE (256 * sizeof(struct entry*))

static struct entry ***** root_hash = NULL;
static struct entry * host_list = NULL;
static unsigned int used_memory = 0;

extern int verbose_dump;

u_int32_t inet_addr(unsigned char*);
unsigned char * inet_ntoa(u_int32_t);

struct entry * get_hostlist(void)
{
	return(host_list);
}

struct ipclassinfo * add_class(char * name, struct ipclassinfo * ci)
{
	char * tmpname = strdup(name);
	char * mask;
	int nmask;
	int i;
	u_int32_t ip;
	u_int32_t netmask;
	struct ipclassinfo *ret;
	
	if(mask = index(tmpname, '/')) 
	nmask = atoi(mask + 1);
	else
	{
		fprintf(stderr, "Class notation: a.b.c.d/mask\n");
		free(tmpname); return(NULL);
	}
	*mask = 0;
	ip = inet_addr(tmpname);
	free(tmpname);
	netmask = 0;
	for(i = 0; i<nmask; i++) netmask = (netmask >> 1) | 0x80000000;
	netmask = htonl(netmask);
#ifdef DEBUG
	fprintf(stderr, "tracing class: %s (%x) ", inet_ntoa(ip), ip);
	fprintf(stderr, "netmask %s (%x)\n", inet_ntoa(netmask), netmask);
#endif
	ret = (struct ipclassinfo*) malloc(sizeof(struct ipclassinfo));
	ret->next = ci;
	ret->net = ip;
	ret->mask = netmask;
	return(ret);
}

int is_ours(struct ipclassinfo * ci, u_int32_t host)
{
	for(; ci; ci = ci->next)
	if((host & ci->mask) == ci->net) return(1);

	return(0);
}

static inline void * bzalloc(int len)
{
	void * ret;
	ret = (void*) malloc(len);
	memset(ret, 0, len);
	used_memory += len;
#ifdef DEBUG
	printf("alloc(%d)\n", len);
#endif
	return(ret);
}

struct entry * get_entry(u_int32_t ip)
{
	struct entry ***** ahash;
	struct entry **** bhash;
	struct entry *** chash;
	struct entry ** dhash;
	struct entry * host;
	int a = ip >> 24;
	int b = (ip >> 16) & 0xff;
	int c = (ip >> 8) & 0xff;
	int d = ip & 0xff;
	if(!(ahash = root_hash)) ahash = root_hash = (struct entry *****) bzalloc(CLASS_SIZE);
	if(!(bhash = ahash[a])) bhash = ahash[a] = (struct entry ****) bzalloc(CLASS_SIZE);
	if(!(chash = bhash[b])) chash = bhash[b] = (struct entry ***) bzalloc(CLASS_SIZE);
	if(!(dhash = chash[c])) dhash = chash[c] = (struct entry **) bzalloc(CLASS_SIZE);
	if(!(host = dhash[d]))
	{
		host = dhash[d] = (struct entry*) bzalloc(sizeof(struct entry));
		host->next = host_list;
		host->ip = ip;
		host_list = host;
	}
	return(host);
}

void add_entry(u_int32_t ip, long long in, long long out)
{
	struct entry * entry = get_entry(ip);
	entry->counters[BAC_BYTES_MOMENTARY_IN]  += in;
	entry->counters[BAC_BYTES_MOMENTARY_OUT] += out;
	entry->counters[BAC_PKT_MOMENTARY_IN]  += in ? 1 : 0;
	entry->counters[BAC_PKT_MOMENTARY_OUT] += out ? 1 : 0;
	entry->counters[BAC_BYTES_OVERALL_IN]  += in;
	entry->counters[BAC_BYTES_OVERALL_OUT] += out;
	entry->counters[BAC_PKT_OVERALL_IN]  += in ? 1 : 0;
	entry->counters[BAC_PKT_OVERALL_OUT] += out ? 1 : 0;
#ifdef DEBUG
//	fprintf(stderr, "Added %Ld, %Ld bytes to entry %s Now counters are %Ld, %Ld\n",  in, out, inet_ntoa(htonl(ip)),entry->input, entry->output);
#endif
}


void zero_entries(void)
{
	unsigned int i, j, k, l;
	struct entry ***** hasha = root_hash;
	if(!hasha) return;
	
	for(i = 0; i<256; i++)
	{
	struct entry **** hashb = hasha[i];
	if(!hashb) continue;
	for(j = 0; j<256; j++)
	{
		struct entry *** hashc = hashb[j];
		if(!hashc) continue;
		for(k = 0; k<256; k++)
		{
		struct entry ** hashd = hashc[k];
		if(!hashd) continue;
		for(l = 0; l<256; l++)
		{
			struct entry * host = hashd[l];
			if(host)
			{
				host->counters[BAC_BYTES_MOMENTARY_IN] = host->counters[BAC_BYTES_MOMENTARY_OUT] =
				host->counters[BAC_PKT_MOMENTARY_IN] = host->counters[BAC_PKT_MOMENTARY_OUT] = 0;
			}
		}
		}
	}
	}
}


				

struct top * findtop(int n, int counter)
{
	unsigned int i, j, k, l, m, c = 0;
	struct top * top = bzalloc(sizeof(struct top));
	struct entry ***** hasha = root_hash;
	
	top -> n = c;
	top -> top = (struct entry**) malloc(sizeof(struct entry*) * n);

	if(!hasha) return(top);
	
	for(i = 0; i<256; i++)
	{
	struct entry **** hashb = hasha[i];
	if(!hashb) continue;
	for(j = 0; j<256; j++)
	{
		struct entry *** hashc = hashb[j];
		if(!hashc) continue;
		for(k = 0; k<256; k++)
		{
		struct entry ** hashd = hashc[k];
		if(!hashd) continue;
		for(l = 0; l<256; l++)
		{
			struct entry * host = hashd[l];
			u_int32_t ip;
			if(host)
			{
			for(m = 0; m<c; m++)
			{
				if(host->counters[counter] > top->top[m]->counters[counter])
				{
					if(m < (n - 1))
					{
						memmove(&top->top[m + 1], &top->top[m], sizeof(struct entry*) * (n - m - 1));
						if(c < n) c++;
					}
					top->top[m] = host;
					break;
				}
			}
			if(c < n && m == c)
			{
				top->top[c++] = host;
			}
			
			}
		}
		}
	}
	}
	top->n = c;
	return(top);
}
