/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <sys/types.h>
#include <stdio.h>
#include <sys/socket.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <net/bpf.h>
#include "lib.h"

#ifndef isblank
#define isblank(c) ((c) == '\t' || (c) == ' ')
#endif

#ifdef SOLARIS_SPARC

inline int inet_my_aton(const char *str, void * addr)
{
	return(inet_pton(AF_INET, str, addr));
}
#else
inline int inet_my_aton(const char *cp, void * addr)
{
	return(inet_aton(cp, addr));
}
#endif

void mkhuman(double value, int max, char * string, char * unit, double base, int h)
{
	char * space = " ";
	if(!unit)
		unit = space = "";

	if(value < base || (!h))
		snprintf(string, max, "%.2f%s %s", value, space, unit);
	else
	{
		value /= base;
		if(value < base)
			snprintf(string, max, "%.2f%sk%s", value, space, unit);
		else
		{
			value /= base;
			if(value < base)
				snprintf(string, max, "%.2f%sM%s", value, space, unit);
			else
			{
				value /= base;
				if(value < base)
					snprintf(string, max, "%.2f%sG%s", value, space, unit);
				else
				{
					value /= base;
					snprintf(string, max, "%.2f%sT%s", value, space, unit);
				}
			}
		}
	}
}

int parseaddr(unsigned char * s, u_int32_t * ip, u_int32_t * netmask)
{
	unsigned char * z;
	unsigned char * e;
	struct in_addr in;
	
	if((z = index(s, '/')))
		*z = 0;

//	printf("Checkout: %s\n", s);
	if(inet_my_aton(s, &in) == 0) return(-1);

	*ip = in.s_addr;

	if(z)
	{
		int masksize;
		z++;
		masksize = strtol(z, (char**) &e, 0);
		if(*e == 0)
		{
			u_int32_t m = 0;

//			printf("Mask: %d\n", masksize);
			if(masksize < 0 || masksize > 32) return(-1);
			while(masksize--)
			{
				m >>= 1;
				m |= 0x80000000;
			}

			*netmask = htonl(m);
		}
		else
		{
			if(!inet_my_aton(z, &in)) return(-1);
			*netmask = in.s_addr;
		}
		
	}
	else
		*netmask = 0xffffffff;
	
	return(0);
}

void freewords(unsigned char ** words)
{
	int i;
	if(words)
		for(i = 0; words[i]; i++) free(words[i]);

	free(words);
}

int wc(unsigned char ** words)
{
	int i;
	if(!words) return(0);
	for(i = 0; words[i]; i++);
	return(i);
}

unsigned char ** explode(unsigned char * line)
{
	unsigned char ** words = NULL;
	
	unsigned char * start;
	unsigned int marker = 0;
	
	unsigned char quotemode = 0;
	
	int n = 0;
	
//	printf("New explode.\n");
	
	
	start = line;
	marker = 0;
	while(*start)
	{
		unsigned char * word;
		int pos = 0;
		
		while(isblank(start[marker])) marker++;	/* skip whitespaces */
		if(! start[marker]) break;
		
		start += marker;
		
		if(start[0] == '"' || start[0] == '\'') quotemode = start[0]; else quotemode = ' ';

		word = (unsigned char*) malloc(strlen(start) + 1);
		
//		printf("Fetching word, quoting mode: '%c' (%c)\n", quotemode, start[0]);
		
		if(quotemode == ' ')
		{
			word[pos++] = start[0];
			start++;
			if(!*start)
			{
				word[pos++] = 0;
				word = (unsigned char*) realloc(word, pos);
				words = (unsigned char**) realloc(words, (n + 1) * sizeof(unsigned char*));
				words[n++] = word;
				break;
			}
//			printf("Proceeding...\n");
		}
		else
		{
			start++;
			if(!start)
			{
				free(word);
				break;
			}
		}
		
		for(marker = 0; start[marker] && start[marker] != quotemode && (quotemode != ' ' || !isblank(start[marker])); marker++)
		{
			if(quotemode == '"')
			{
				if(start[marker] == '\\')
				{
					unsigned char c;
					marker++;

					if(!start[marker]) break;
					switch(start[marker])
					{
						case 'n': c = '\n'; break;
						case 'r': c = '\r'; break;
						case 't': c = '\t'; break;
						default: c = start[marker]; break;
					}
					word[pos++] = c;
				}
				else
					word[pos++] = start[marker];
			}
			else
				word[pos++] = start[marker];
//			printf("character[%d]: %c ", pos - 1, start[marker]);
		}
		
//		printf("\n. START: '%s' MARKER: '%d'", start, marker);
		start += marker;
		if(*start) start++;
		
		word[pos++] = 0;
		word = (unsigned char*) realloc(word, pos);
		words = (unsigned char**) realloc(words, (n + 1) * sizeof(unsigned char*));
		words[n++] = word;

		marker = 0;
	}
	if(words)
	{
		words = (unsigned char**) realloc(words, (n + 1) * sizeof(unsigned char*));
		words[n++] = NULL;
	}
//	printf("Returning words:\n");
	return(words);
}

int human2machine(unsigned char * val, double * v)
{
	unsigned char *endptr;
	double c;
	double mult = 0;
	
	c = strtod(val, (char**) &endptr);
	if(*endptr == *val) return -1;
	if(!*endptr)
	{
		*v = c;
		return(0);
	}
	if(strcmp(endptr, "pps") == 0 || strcmp(endptr, "bps") == 0) mult = 1;
	if(!strcmp(endptr, "kpps")) mult = 1000;
	if(!strcmp(endptr, "Mpps")) mult = 1000000;
	if(!strcmp(endptr, "Gpps")) mult = 1000000000;

	if(!strcmp(endptr, "kbps")) mult = 1024 / 8;
	if(!strcmp(endptr, "Kbps")) mult = 1024 / 8;
	if(!strcmp(endptr, "Mbps")) mult = 1024 * 1024 / 8;
	if(!strcmp(endptr, "Gbps")) { mult = 1024 * 1024 * 1024; mult /= 8; }

	if(!strcmp(endptr, "kBps")) mult = 1024;
	if(!strcmp(endptr, "KBps")) mult = 1024;
	if(!strcmp(endptr, "MBps")) mult = 1024 * 1024;
	if(!strcmp(endptr, "GBps")) mult = 1024 * 1024 * 1024;
	
	if(mult == 0) return(-1);
	c *= mult;
	*v = c;
	return(0);
}

#ifdef AUTOSENSE_DLH
int llheader(int datalink)
{
	switch(datalink)
	{
		case DLT_EN10MB:	return(14);
		case DLT_IEEE802:	return(22);
		case DLT_FDDI:		return(21);
		case DLT_SLIP:		return(16);
		case DLT_PPP:		return(4);
		case DLT_RAW:		return(0);
		case DLT_NULL:		return(4);
	}
	fprintf(stderr, "WARNING: Link layer header length autosense FAILED! Using default '14' (Ethernet), PLEASE READ the -p option description in the manual...\n");
	return(14);
}
#endif

int gethost(unsigned char * host, u_int32_t *ip)
{
	struct hostent *he;
	struct in_addr sin;
	int h = 0;
	
	if ((he = gethostbyname(host)) != NULL)
	{
		memcpy(&sin, he->h_addr_list[0], he->h_length);
		h = 1;
	}
	else
	{
		if(inet_my_aton(host, &sin)) h = 1;
	}		
	*ip = sin.s_addr;
	return(h);
}
							    
