/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/timeb.h>
#include <signal.h>
#include <time.h>
#include <pcap.h>
#include "lib.h"
#ifdef HAVE_REGEX
#include <regex.h>
#endif
#include "iphash.h"
#include "protos.h"
#include "colors.h"
#include "streams.h"
#include "streamstat.h"

extern unsigned char * opt_bpfexpression;
extern int opt_human;
extern int opt_dns;
extern int opt_cols;
extern int opt_cls;

static int opt_bits = 0;
static int opt_frames = 1;
static struct ipclassinfo * ipci = NULL;

static int opt_printtop = 10;
static int opt_local = 0;
static int opt_dump = 0;
extern int opt_color;
static int opt_totals = 0;
static int opt_counter = BAC_BYTES_MOMENTARY_IN;

static void host_inetntoa(char * str, int max, u_int32_t ip)
{
	if(opt_dns)
		snprintf(str, max, "%s", dns_lookup(ip));
	else
		snprintf(str, max, "%d.%d.%d.%d", (ip >> 24) & 0xff, (ip >> 16) & 0xff, (ip >> 8) & 0xff, ip & 0xff);
}

static void output(struct timeb * now, int miliFromStart, int miliFromLast, 
					unsigned long long totalbytes, unsigned long long totalpackets,
					unsigned long nowbytes, unsigned long nowpackets)
{
	struct top * top;


	if(opt_dump)
	{
		struct entry * en;
		char buffer[30];
		
		for(en = get_hostlist(); en; en = en->next)
		{
			int i;
			host_inetntoa(buffer, 30, en->ip);
			printf("%s", buffer);
			for(i = 0; i<8; i++) printf(":%Ld", en->counters[i] * ((opt_bits && (i < 4)) ? 8 : 1));
			printf("\n");
		}

	}
	else
	{
		int j;
		char frame = (opt_frames) ? '|' : ' ';
		char indent[100];
		
		top = findtop(opt_printtop, opt_counter);
	
		if(top->n > 0)
		{           
//            213.251.115.111  123.43M 123.53k 123.63M 126.73M 112.33M 443.32M 122.34M 122.33M

			if(opt_color) printf("%s", COLOR_AB_HEADER);
			
			snprintf(indent, 100, customize_format(opt_cols * 2, 65 * 2, "%%-%d.%ds"), "");
			
			
			if(opt_totals)
			{
				printf("%s   Momentary Rx    Momentary Tx      Overall Rx      Overall Tx \n", indent);
				if(opt_bits)
					printf("%s    bps     pps %c   bps     pps %c  bits    pkts %c  bits    pkts \n", indent, frame, frame, frame);
				else
					printf("%s    Bps     pps %c   Bps     pps %c bytes    pkts %c bytes    pkts \n", indent, frame, frame, frame);
			}
			else
			{
				printf("%s   Momentary Rx    Momentary Tx      Average Rx      Average Tx \n", indent);
				if(opt_bits)
					printf("%s    bps     pps %c   bps     pps %c   bps     pps %c   bps     pps \n", indent, frame, frame, frame);
				else
					printf("%s    Bps     pps %c   Bps     pps %c   Bps     pps %c   Bps     pps \n", indent, frame, frame, frame);
			}
			
			if(opt_color) printf("\033[0m");


			for(j = 0; j<top->n; j++)
			{
				static char sstr[8][20];
				char ip[50];
				int k;
				
				host_inetntoa(ip, 50, top->top[j]->ip);
				for(k = 0; k<8; k++)
				{
					double v;
					v = top->top[j]->counters[k];
					
					if((k & 2) || (!opt_totals))
					{
						if(k & 2)
							v /= miliFromLast;
						else
							v /= miliFromStart;
							
						v *= 1000;
					}
					if(opt_bits) v *= 8;
					
					mkhuman(v, 20, sstr[k], NULL, k > 3 ? 1000 : 1024, opt_human);
				}
				printf(customize_format(opt_cols * 2, 65 * 2, "%%s%%-%d.%ds %%s%%7s %%s%%7s%%c%%s%%7s %%s%%7s%%c%%s%%7s %%s%%7s%%c%%s%%7s %%s%%7s%%s\n"),
						CM_AB_HOST(opt_color),
						ip,
						CM_AB_BMI(opt_color), sstr[BAC_BYTES_MOMENTARY_IN],
						CM_AB_PMI(opt_color), sstr[BAC_PKT_MOMENTARY_IN], frame,
						CM_AB_BMO(opt_color), sstr[BAC_BYTES_MOMENTARY_OUT],
						CM_AB_PMO(opt_color), sstr[BAC_PKT_MOMENTARY_OUT], frame,
						CM_AB_BOI(opt_color), sstr[BAC_BYTES_OVERALL_IN],
						CM_AB_POI(opt_color), sstr[BAC_PKT_OVERALL_IN], frame,
						CM_AB_BOO(opt_color), sstr[BAC_BYTES_OVERALL_OUT],
						CM_AB_POO(opt_color), sstr[BAC_PKT_OVERALL_OUT],
						opt_cls ? ("\033[K") : (""));
			}
		}

		free(top->top); free(top);
	}
	zero_entries();
}


static void help2(void)
{
	printf("\n\nABUSERS USAGE: ... -A -c <class>/<prefixlen> [-c <class>/<prefixlen> [-c <class>/<prefixlen> ...]] [-a n] [-bvtld] [-O | -M] [-P | -B] [-T | -R] [bpf filter]\n");
	printf("  -b     Bit units (show output in bits/sec instead of bytes/sec)\n");
	printf("  -f     Print spaces instead of column separators (|).\n");
	printf("  -t     Print total amount of transferred data instead of overall speed.\n");
	printf("  -a <n> Print n top hosts (default: 10).\n");
	printf("  -c <c> Take nodes from class c into consideration.\n");
	printf("  -l     Account local transfers too.\n");
	printf("  -d     Alternative output format.\n");
	printf("  -O     Order by overall rates.\n");
	printf("  -M     Order by momentary rates (default).\n");
	printf("  -P     Order by packet counters.\n");
	printf("  -B     Order by byte counters (default).\n");
	printf("  -T     Order by output counters.\n");
	printf("  -R     Order by input counters (default).\n");
}

static void counter(const unsigned char * data, int caplen, int len)
{
	struct ip *iph;
	u_int32_t sip;
	u_int32_t dip;
	u_int32_t host;
	long long in = 0, out = 0;

	if(caplen < sizeof(struct ip)) return;

	iph = (struct ip*) data;
	if(iph->ip_v != 4) return;

	sip = iph->ip_src.s_addr;
	dip = iph->ip_dst.s_addr;

 	if(is_ours(ipci, sip)) out = len;
	if(is_ours(ipci, dip)) in = len;

	if(!(in || out)) return;
	if(in && out && (!opt_local)) return;

	if(in)  add_entry(ntohl(dip), in, 0);
	if(out) add_entry(ntohl(sip), 0, out);
}

void r_abusers_setup(int argc, char ** argv,
						void (**custom_output)(struct timeb*, int, int, unsigned long long,
							  unsigned long long, unsigned long, unsigned long),
						void (**custom_counter)(const unsigned char *, int, int))
{
	int c;
	
	*custom_output = output;
	*custom_counter = counter;
	while((c = getopt(argc, argv, "?hba:c:ldtOMPBTR")) != EOF)
	{
		switch(c)
		{
			case '?':
			case 'h':
				help(argv[0]);
				help2();
				exit(0);
			case 'b': opt_bits = 1; break;
			case 'a':
				opt_printtop = atoi(optarg);
				if(opt_printtop < 1) opt_printtop = 1;
				break;
			case 'c':
				ipci = add_class(optarg, ipci);
				if(!ipci) exit(1);
				break;
			case 'd': opt_dump = 1; break;
			case 't': opt_totals = 1; break;
			case 'l': opt_local = 1; break;
			case 'R': opt_counter &= ~1; break;
			case 'T': opt_counter |=  1; break;
			case 'O': opt_counter &= ~2; break;
			case 'M': opt_counter |=  2; break;
			case 'B': opt_counter &= ~4; break;
			case 'P': opt_counter |=  4; break;
		}
	}

	if(!ipci)
	{
		fprintf(stderr, "No IP classes specified! The -c option is required in this mode. Consult %s -A -h.\n", argv[0]);
		exit(1);
	}

	if(argv[optind])
	{
		if(opt_bpfexpression)
		{
			fprintf(stderr, "BPF Filter expression already specified: '%s' and '%s'...\n", opt_bpfexpression, argv[optind]);
			exit(1);
		}
		else
			opt_bpfexpression = argv[optind++];
	}
}
