/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/timeb.h>
#include <signal.h>
#include <time.h>
#include <pcap.h>
#include "lib.h"
#ifdef HAVE_REGEX
#include <regex.h>
#endif
#include "iphash.h"
#include "protos.h"

extern unsigned char * opt_bpfexpression;

static int opt_bits = 0;
extern int opt_human;

static void output(struct timeb * now, int miliFromStart, int miliFromLast, 
					unsigned long long totalbytes, unsigned long long totalpackets,
					unsigned long nowbytes, unsigned long nowpackets)
{
	double bpsFromStart, bpsFromLast, ppsFromStart, ppsFromLast;
	char bpsFromLastStr[100];
	char bpsFromStartStr[100];
	char ppsFromLastStr[100];
	char ppsFromStartStr[100];

	ppsFromStart = totalpackets; ppsFromStart /= miliFromStart;
	ppsFromLast = nowpackets; ppsFromLast /= miliFromLast;
	bpsFromStart = totalbytes; bpsFromStart /= miliFromStart;
	bpsFromLast = nowbytes; bpsFromLast /= miliFromLast;
	ppsFromStart *= 1000; ppsFromLast *= 1000;
	bpsFromStart *= 1000; bpsFromLast *= 1000;
	
	if(opt_bits)
		mkhuman(bpsFromLast * 8, 100, bpsFromLastStr, "bps", 1024, opt_human);
	else
		mkhuman(bpsFromLast, 100, bpsFromLastStr, "Bps", 1024, opt_human);
		
	mkhuman(ppsFromLast, 100, ppsFromLastStr, "pps", 1000, opt_human);
	
	if(opt_bits)
		mkhuman(bpsFromStart * 8, 100, bpsFromStartStr, "bps", 1024, opt_human);
	else
		mkhuman(bpsFromStart, 100, bpsFromStartStr, "Bps", 1024, opt_human);

	mkhuman(ppsFromStart, 100, ppsFromStartStr, "pps", 1000, opt_human);
	printf("=> Currently %s/%s, Average: %s/%s\n", bpsFromLastStr, ppsFromLastStr, bpsFromStartStr, ppsFromStartStr);
}


static void help2(void)
{
	printf("\n\nRATE ESTIMATOR USAGE: ... -R [-b] [bpf filter]\n");
	printf("  -b   Bit units (show output in bits/sec instead of bytes/sec)\n");
}

static void counter(const unsigned char * data, int caplen, int len)
{
}

void r_estimator_setup(int argc, char ** argv,
						void (**custom_output)(struct timeb*, int, int, unsigned long long,
							  unsigned long long, unsigned long, unsigned long),
						void (**custom_counter)(const unsigned char *, int, int))
{
	int c;
	
	*custom_output = output;
	*custom_counter = counter;
	while((c = getopt(argc, argv, "?hb")) != EOF)
	{
		switch(c)
		{
			case '?':
			case 'h':
				help(argv[0]);
				help2();
				exit(0);
			case 'b': opt_bits = 1; break;
		}
	}
	if(argv[optind])
	{
		if(opt_bpfexpression)
		{
			fprintf(stderr, "BPF Filter expression already specified: '%s' and '%s'...\n", opt_bpfexpression, argv[optind]);
			exit(1);
		}
		else
			opt_bpfexpression = argv[optind++];
	}
}
