/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef STREAM_ANALYZER
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/timeb.h>
#include <signal.h>
#include <time.h>
#include <pcap.h>
#include "lib.h"
#ifdef HAVE_REGEX
#include <regex.h>
#endif
#include "iphash.h"
#include "protos.h"
#include "sta.h"
#include "streams.h"
#include "streamstat.h"
#include "colors.h"
#include "process.h"

extern int opt_caplen;
extern int opt_cls;
extern int opt_human;
extern int opt_cols;
extern int opt_dns;

static int opt_memlimit = 0;
static int opt_memfree = 4000;
static int opt_verbose = 0;

static int opt_top = 5;
static int opt_brief = 0;
static int opt_mode = COUNTER_BYTE;
static int opt_counter = LRECV;
static int opt_sort = ORDER_ASC;
extern int opt_color;
static int opt_spaces = 0;
static int opt_overall = 0;
static int opt_resources = 0;
static char opt_opmode = 0;

static void output(struct timeb * now, int miliFromStart, int miliFromLast, 
					unsigned long long totalbytes, unsigned long long totalpackets,
					unsigned long nowbytes, unsigned long nowpackets)
{
	struct itimerval itv;
	static double time_real = 0, time_this = 0; /* in 1/1000 s */

	int mem, sc, ssc, hc;

	getitimer(ITIMER_PROF, &itv);
	
	time_real = miliFromStart;
	time_this = (PROFILER_STARTUP - itv.it_value.tv_sec) * 1000 + ((0 - itv.it_value.tv_usec) / 1000);
	
	if(time_real)
		time_this /= time_real / 100;
	else
		time_this = 0;
	
	mem = stastream_streamcount() * sizeof(sta_stream) + stastream_hostcount() * sizeof(sta_host);
	mem /= 1024;

	if(opt_memlimit)
	{
		if(mem > opt_memlimit)
			stastream_cleanstreams(now, opt_memfree);
	}

	
	stastream_expire(now);

	sc = stastream_streamcount();
	ssc = stastream_stalestreamcount();
	hc = stastream_hostcount();

	if(opt_resources)
	{
		printf(opt_color ?
		(COLOR_RESOURCEREPORT "### %d streams (%d stale) %d hosts, hash tables: %d kB, CPU usage %.0f %%\033[0m  \n") :
		("### %d streams (%d stale) %d hosts, hash tables: %d kB, CPU usage %.0f %%  \n"), sc, ssc, hc, mem, time_this);
	}
	
	switch(opt_opmode)
	{
		case 'n':
		{
			sta_host ** list;
			int n, i;


			list = stastat_gethostlist(opt_top, &n, opt_mode, opt_counter, opt_sort, 0);
			if((!opt_brief) && n)
				printf("%s\n", stastat_streamheader("  ", opt_cols, opt_verbose, opt_color));

			for(i = 0; i<n; i++)
			{
				struct in_addr in;
				sta_stream ** streams;
				sta_host * h = list[i];
				int n, j;
				char overall_bytes[20];
				char overall_packets[20];
				char overall_bps[20];
				char overall_pps[20];
				char host[50];
			
	
				if(opt_dns)
					snprintf(host, 50, "%s", dns_lookup(ntohl(h->ip)));
				else
				{
					in.s_addr = h->ip;
					snprintf(host, 50, "%s", inet_ntoa(in));
				}
	
				if(opt_counter == LRECV || opt_counter == RECV)
				{
					mkhuman(h->bytes[RECV], 20, overall_bytes, "B", 1024, opt_human);
					mkhuman(h->packets[RECV], 20, overall_packets, "p", 1000, opt_human);
					mkhuman(h->bytes[LRECV] * 1000.0 / miliFromLast, 20, overall_bps, "Bps", 1024, opt_human);
					mkhuman(h->packets[LRECV] * 1000.0 / miliFromLast, 20, overall_pps, "pps", 1000, opt_human);
				}
				else
				{
					mkhuman(h->bytes[SENT], 20, overall_bytes, "B", 1024, opt_human);
					mkhuman(h->packets[SENT], 20, overall_packets, "p", 1000, opt_human);
					mkhuman(h->bytes[SENT] * 1000.0 / miliFromLast, 20, overall_bps, "Bps", 1024, opt_human);
					mkhuman(h->packets[SENT] * 1000.0 / miliFromLast, 20, overall_pps, "pps", 1000, opt_human);
				}

// ### 01.  123 123.45k/123.45k 123.45k/123.45k
// ### 02.  6s  7.67 kB/115.00  p 683.00  Bps/10.00  pps
				printf(customize_format(opt_cols * 2, 54 * 2, 
						"%%s### %%.2d.%%s %%s%%-%d.%ds%%s %%s%%3ds%%s %%s%%8s/%%8s%%s %%s%%8s/%%8s%%s%%s\n"),
						CM_POSITION(opt_color),
						i + 1,
						CM_PLAIN(0, opt_color),
						CM_HOST(0, opt_color),
						host,
						CM_PLAIN(0, opt_color),
						CM_STREAMCOUNT(opt_color),
						list[i]->streamcount,
						CM_PLAIN(0, opt_color),
						CM_DATA(0, opt_color),
						overall_bytes,
						overall_packets,
						CM_PLAIN(0, opt_color),
						CM_RATE(0, opt_color),
						overall_bps,
						overall_pps,
						CM_PLAIN(0, opt_color),
						opt_cls ? "\033[K" : "");
				
				if(!opt_brief)
				{
					streams = stastat_getstreams(list[i], &n);
					stastat_sortstreams(streams, n);

					for(j = 0; j<n; j++)
					{
						char * sd;
						if(streams[j]->properties & SPROP_CLOSED) continue;
						sd = stastat_stream2a("  ", streams[j], opt_cols, opt_verbose, opt_human, opt_color, opt_spaces, miliFromLast);
						printf("  %s%s\n", sd, opt_cls ? "\033[K" : "");
						free(sd);
					}
					free(streams);
					printf("%s\n", opt_cls ? "\033[K" : "");
				}
			}
			free(list);
			break;
		}
		case 's':
		{
			sta_stream ** list;
			int n, i;

			list = stastat_getstreamlist(opt_top, &n, opt_mode, opt_overall, opt_sort, 0);
			if(n)
				printf("%s\n", stastat_streamheader("   ", opt_cols, opt_verbose, opt_color));
			for(i = 0; i<n; i++)
			{
				char * sd;
				sd = stastat_stream2a("   ", list[i], opt_cols, opt_verbose, opt_human, opt_color, opt_spaces, miliFromLast);
				printf((opt_color) ? (COLOR_POSITION "%2d \033[0m%s\n") : 
									 "%2d %s%s\n", i + 1, sd, opt_cls ? "\033[K" : "");
				free(sd);
			}
			free(list);
			break;
		}
		case 'p':
		{
			struct protoinfo ** list;
			int n, i;

			list = stastat_getprotolist(opt_top, &n, opt_mode, opt_overall, opt_sort);
			if((!opt_brief) && n)
				printf("%s\n", stastat_streamheader("  ", opt_cols, opt_verbose, opt_color));
			for(i = 0; i<n; i++)
			{
				char overall_bytes[20];
				char overall_packets[20];
				char overall_bps[20];
				char overall_pps[20];
				struct protoinfo * pi = list[i];

				mkhuman(pi->counters[PC_OVERALL_BYTES], 20, overall_bytes, "B", 1024, opt_human);
				mkhuman(pi->counters[PC_OVERALL_PACKETS], 20, overall_packets, "p", 1000, opt_human);
				mkhuman(pi->counters[PC_LAST_BYTES] * 1000.0 / miliFromLast, 20, overall_bps, "Bps", 1024, opt_human);
				mkhuman(pi->counters[PC_LAST_PACKETS] * 1000.0 / miliFromLast, 20, overall_pps, "pps", 1000, opt_human);

				printf(opt_color ?
					(COLOR_POSITION "### %2d. \033[0m" COLOR_PROTOCOL "%s \033[0m" COLOR_DATA "%s/%s\033[0m " COLOR_RATE "%s/%s\033[0m%s\n") :
					("### %.2d. %s %s/%s %s/%s%s\n"),
					i + 1, sta_getprotoname(pi->protoid), 
					overall_bytes, overall_packets, overall_bps, overall_pps, opt_cls ? "\033[K" : "");

				if(!opt_brief)
				{
					sta_stream ** streams;
					sta_stream * streamlist;
					sta_stream * current;
					int n;
					
					streamlist = stastream_getprotolist(pi->protoid);
					for(n = 0, current = streamlist; current; current = current->proto_next) n++;
					
					if(n)
					{
						int j;
						
						streams = (sta_stream**) sta_malloc(sizeof(sta_stream*) * n);
						for(n = 0, current = streamlist; current; current = current->proto_next) 
							streams[n++] = current;
										
						stastat_sortstreams(streams, n);

						for(j = 0; j<n; j++)
						{
							char * sd;
							if(streams[j]->properties & SPROP_CLOSED) continue;
							sd = stastat_stream2a("  ", streams[j], opt_cols, opt_verbose, opt_human, opt_color, opt_spaces, miliFromLast);
							printf("  %s%s\n", sd, opt_cls ? "\033[K" : "");
							free(sd);
						}
						free(streams);
						printf("%s\n", opt_cls ? "\033[K" : "");
					}
				}
// 00 U xxx.yyy.zzz.ttt:ppppp xxx.yyy.zzz.ttt:ppppp kkk.kkK ppp.ppK kps.kkK bps.ppK

//				char * sd;
//				
//				sd = stastat_stream2a("     ", list[i], opt_verbose, opt_human, opt_color, opt_tab, miliFromLast);
//				printf((opt_color) ? (COLOR_POSITION "## %2d. \033[0m%s\n") : 
//									 "## %2d. %s\n", i + 1, sd);
//				free(sd);
			}
			free(list);
			break;
		}

	}
	stastream_clear();
	
}

static void help2(void)
{
	printf("\n\nSTREAM TRAFFIC ANALYZER USAGE: ... -T [-uvrct] [-m <ml> [-f <f>]] (-n | -s | -p) <n> [-R | -S] [-M | -O] [-B | -P] [-A | -D] [-b]\n");
	printf("  -v     Increase verbosity level.\n");
	printf("  -r     Output resource statistics (CPU usage, memory used by hash tables).\n");
	printf("  -t     Separate output columns by tabs.\n");
	printf("  -m <m> Limit memory used by stream and host data to m kilobytes.\n");
	printf("  -d <f> After a memory overlimit, dispose f stream entries (default: 4000).\n");
	printf("  -f     Print spaces instead of standard column separators (|).\n");
	printf("  -n <n> Node-driven reports - report n top nodes:\n");
	printf("      -R On received data (default).\n");
	printf("      -S On sent data.\n");
	printf("      -M On momentary rates (default).\n");
	printf("      -O On overall rates.\n");
	printf("      -B Counting bytes (default).\n");
	printf("      -P Counting packets.\n");
	printf("      -A Ascending sort.\n");
	printf("      -D Descending sort (default).\n");
	printf("      -b Brief output - do not print stream list for each host.\n");
	printf("  -s <n> Stream-driven reports - report n top streams:\n");
	printf("      -B On bytes transferred within a stream (default).\n");
	printf("      -P On packets transferred within a stream.\n");
	printf("      -M On momentary rates (default).\n");
	printf("      -O On overall rates.\n");
	printf("      -A Ascending sort.\n");
	printf("      -D Descending sort (default).\n");
	printf("  -p <n> Protocol-driven reports - report n top protocols:\n");
	printf("      -B On bytes transferred using a protocol (default).\n");
	printf("      -P On packets transferred using a protocol.\n");
	printf("      -M On momentary rates (default).\n");
	printf("      -O On overall rates.\n");
	printf("      -A Ascending sort.\n");
	printf("      -D Descending sort (default).\n");
	printf("      -b Brief output - do not print stream list for each protocol.\n");
}

static void counter(const unsigned char * data, int caplen, int len)
{
	struct timeval tv;
	struct timezone tz;

	gettimeofday(&tv, &tz);
	stastream_classify(data, caplen, len, &tv);
}

void r_stream_setup(int argc, char ** argv,
						void (**custom_output)(struct timeb*, int, int, unsigned long long,
							  unsigned long long, unsigned long, unsigned long),
						void (**custom_counter)(const unsigned char *, int, int))
{
	int c, dir = 0, mom = 0;
	
	*custom_output = output;
	*custom_counter = counter;
	while((c = getopt(argc, argv, "?hvd:m:fn:s:p:RSMOBPADbrt")) != EOF)
	{
		switch(c)
		{
			case '?':
			case 'h':
				help(argv[0]);
				help2();
				exit(0);
			case 'v': opt_verbose++; break;
			case 'm':
				opt_memlimit = atoi(optarg);
				break;
			case 'd':
				opt_memfree = atoi(optarg);
				break;
			case 'n':
			case 's':
			case 'p':
				if(opt_opmode)
				{
					fprintf(stderr, "You have to select only one of: -n, -s or -p. Consult %s -T -h\n", argv[0]);
					exit(1);
				}
				opt_opmode = c;
				opt_top = atoi(optarg);
				if(opt_top < 1) opt_top = 1;
				break;
			case 'A': opt_sort = ORDER_DESC; break;
			case 'D': opt_sort = ORDER_ASC; break;
			case 'B': opt_mode = COUNTER_BYTE; break;
			case 'P': opt_mode = COUNTER_PACKET; break;
			case 'R': dir = 1; break;
			case 'S': dir = 2; break;
			case 'M': mom = 1; break;
			case 'O': mom = 2; break;
			case 'f': opt_spaces = 1; break;
			case 'r': opt_resources = 1; break;
			case 'b': opt_brief = 1; break;
		}
	}
	
	if(!opt_opmode)
	{
		opt_opmode = 'n';
		opt_top = 10;
	}
	
	if(opt_opmode != 'n' && dir)
	{
		fprintf(stderr, "The -R and -S options do not make sense in this mode. Consult %s -T -h.\n", argv[0]);
		exit(1);
	}

	if(opt_opmode == 'n')
	{	
		if(dir && (!mom)) mom = 1;
		if(mom && (!dir)) dir = 1;

		if(dir == 1 && mom == 1) opt_counter = LRECV;
		if(dir == 2 && mom == 1) opt_counter = LSENT;
		if(dir == 1 && mom == 2) opt_counter = RECV;
		if(dir == 2 && mom == 2) opt_counter = SENT;
	}
	else
		if(mom == 2) opt_overall = 1;
	
	if(opt_caplen < 1500) opt_caplen = 1500;
}
#endif
