/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef HAVE_REGEX
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/timeb.h>
#include <signal.h>
#include <time.h>
#include <pcap.h>
#include "lib.h"
#ifdef HAVE_REGEX
#include <regex.h>
#endif
#include "iphash.h"

static regex_t global_regex_c;
static char * regex_checking_buffer = NULL;
extern int opt_caplen;


char* unescape(char *string, int *length, unsigned char terminator)
{
	char *out;
	char c, d, e;
	int i, k;

	out = (char*) malloc(strlen(string)); k=0; /* in the worst case we'll need so much */
	for(i = 0; string[i] != 0; i++) /* foreach character in string */
	{
		if(string[i] == '\\') /* is it '\' ? */
		{
			c = string[i + 1]; /* get next character .. */
			if(!c) continue; /* if it's end of string, forget that */
			if(c == 'x') /* x - means hexadecimal code of character */
			{
				if((d = string[i + 2])) /* get first hex digit */
				{
					if((e = string[i + 3])) /* get second hex digit */
					{
						if(d >= 'a' && d <= 'f') d = d - 'a' + 10; else d -= '0';
						if(e >= 'a' && e <= 'f') e = e - 'a' + 10; else e -= '0';
						out[k++] = d << 4 | e; /* calculate character code and write into final string */
						i += 3; /* x<hex_digit><hex_digit> */
						continue; 
					}
				}
			}
			if(c == 'n')
			{
				out[k++] = '\n';
				i += 1;
				continue;
			}
			if(c == 'r')
			{
				out[k++] = '\r';
				i += 1;
				continue;
			}
			if(c == 't')
			{
				out[k++] = '\t';
				i += 1;
				continue;
			}
			if(c == '\\')
			{
				out[k++] = c;
				i += 1;
				continue;
			}
			out[k++] = string[i];
			out[k++] = c;
			i += 1;
		}
		else
		{
			if(string[i] == terminator) break; /* we reached end of string */
			out[k++] = string[i]; /* add that end marker to final string */
		}
	}
	out[k] = 0;
	if(length) *length = i;
	return(out);
}


void prepare_buffer(const char * data, int len, char nul_char)
{
	int safe_len = (len > opt_caplen) ? opt_caplen : len;
	int i;
	for(i = 0; i<safe_len; i++)
	{
		char c;
		c = data[i];
		if(c == 0) c = nul_char;
		regex_checking_buffer[i] = c;
	}
	regex_checking_buffer[i] = 0;
}

int check_regex(const char * data, int len, char nul_char)
{
	prepare_buffer(data, len, nul_char);
	if(regexec(&global_regex_c, regex_checking_buffer, 0, NULL, 0) == 0)
		return(1);
	else
		return(0);
	
}
void prepare_regex(char *regex, int caplen)
{
	if(regex)
	{
		int rc;
		
		regex = unescape(regex, NULL, 0);
		rc = regcomp(&global_regex_c, regex, REG_EXTENDED | REG_NOSUB);
		if(rc)
		{
			char errbuff[100];
			regerror(rc, &global_regex_c, errbuff, 100);
			fprintf(stderr, "regcomp('%s'): %s\n", regex, errbuff);
			exit(1);
		}
		regex_checking_buffer = (char*) malloc(caplen + 1);
		if(!regex_checking_buffer)
		{
			fprintf(stderr, "Unable to allocate regex checking buffer, out of memory?\n");
			exit(1);
		}
	}
}




static regex_t *global_extractor_c = NULL;
static char ** stringtab = NULL;
static int *lentab = NULL;
static int ext_count = 0;

static char ** formattab = NULL;
static int * ordertab = NULL;
int ordercount = 0;


void prepare_extractor(char *regex, char * order, int caplen)
{
	int state = '(';
	char * pos, * nx;

	nx = pos = regex = unescape(regex, NULL, 0);
	
	for(; nx;)
	{
		char * part;
		int rc;

		for(nx = index(pos, state); nx; nx = index(nx + 1, state))
			if(nx != regex)
			{
				if((nx - 1) != regex)
					if(*(nx - 2) == '\\') continue;
				
				if(*(nx - 1) == '\\')
					break;
			}
		
		if(nx)
		{
			*(nx - 1) = 0;
			*(nx++) = 0;
		}
		
		if(state == '(')
			state = ')';
		else
			state = '(';
		
		part = strdup(pos);
		global_extractor_c = (regex_t*) realloc(global_extractor_c, sizeof(regex_t) * (ext_count + 1));
		{
			rc = regcomp(&global_extractor_c[ext_count], part, REG_EXTENDED);
			if(rc)
			{
				char errbuff[100];
				regerror(rc, &global_extractor_c[ext_count], errbuff, 100);
				fprintf(stderr, "regcomp('%s'): %s\n", part, errbuff);
				exit(1);
			}
			ext_count++;
		}
		pos = nx;
	}
	
	if(ext_count % 2 == 0)
	{
		fprintf(stderr, "Unmatched '(' or ')' (%d sub-expressions)\n", ext_count);
		exit(1);
	}
	if(!ext_count)
	{
		fprintf(stderr, "Empty extractor match regex.\n");
		exit(1);
	}

	regex_checking_buffer = (char*) malloc(caplen + 1);

	if(!regex_checking_buffer)
	{
		fprintf(stderr, "Unable to allocate regex checking buffer, out of memory?\n");
		exit(1);
	}
	stringtab = (char**) malloc(sizeof(char*) * ext_count);
	lentab = (int*) malloc(sizeof(int) * ext_count);

	if(order)
	{
		nx = pos = order;
	
		for(; nx;)
		{
			char * part;

			nx = index(pos, '\\');
		
			if(nx)
				*(nx++) = 0;
	
			formattab = (char**) realloc(formattab, (ordercount + 1) * sizeof(char*));
			ordertab = (int*) realloc(ordertab, (ordercount + 1) * sizeof(int));
				
			formattab[ordercount] = strdup(pos);
			if(nx)
			{
				if(*nx)
				{
					ordertab[ordercount] = *(nx) - '1';
					if(ordertab[ordercount] >= (ext_count / 2))
					{
						fprintf(stderr, "Reference %c in format string is invalid.\n", *nx);
						exit(1);
					}
					nx++;
					ordercount++;
				}
			}
			pos = nx;
		}
	}
}

#define MAX_STRSIZE 512

int extractor(const char * data, int len, char nul_char, int showips)
{
	regmatch_t match;
	int lastchar = -1;
	int i, j = 0;
	unsigned char strout[MAX_STRSIZE];
	
	
	prepare_buffer(data, len, nul_char);
	for(i = 0; i < ext_count; i++)
	{
		if(regexec(&global_extractor_c[i], regex_checking_buffer + ((lastchar > 0) ? (lastchar) : 0), 1, &match, 0) != 0) break;
		
		if(i % 2 == 1)
		{
			stringtab[j] = regex_checking_buffer + lastchar + match.rm_so;
			lentab[j++] = match.rm_eo - match.rm_so + 1;
		}
		
		if(lastchar < 0)
			lastchar = match.rm_eo;
		else
		{
			if(match.rm_so != 0) break;
			lastchar += match.rm_eo;
		}
	}
	
	if(i == ext_count)
	{
		if(showips)
		{
			struct in_addr in;
			if(len >= sizeof(struct ip))
			{
				in = ((struct ip*) data)->ip_src;
				printf("%s => ", inet_ntoa(in));
				in = ((struct ip*) data)->ip_dst;
				printf("%s ", inet_ntoa(in));
			}
		}
		if(ordertab)
		{
			int oc;
			for(oc = 0; oc<ordercount; oc++)
			{
				int l, o;
				printf("%s", formattab[oc]);
				o = ordertab[oc];
				
				l = lentab[o];
				if(l > (MAX_STRSIZE - 1)) l = MAX_STRSIZE - 1;
				snprintf(strout, l, "%s", stringtab[o]);
				printf("%s", strout);
			}
			printf("%s\n", formattab[oc]);
		}
		else
		{
			for(i = 0; i < j; i++)
			{
				int l = lentab[i];
				if(l > (MAX_STRSIZE - 1)) l = MAX_STRSIZE - 1;
				snprintf(strout, l, "%s", stringtab[i]);
				printf("%s ", strout);
			}
			printf("\n");
		}
	}
}

#endif
