/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <sys/types.h>
#include <pwd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#define MAX_LINE_SIZE 512

static char ** decode_line(char * line)
{
	int l, i, j = 0, rc = 0;
	char line_translated[MAX_LINE_SIZE];
	char ** ret = NULL;
	char * r;
	
	r = index(line, '\n');
	if(r) *r = 0;

	l = strlen(line);
	
	for(i = 0; i < l; i++)
	{
		if(line[i] == '|')
		{
			line_translated[j++] = 0;
			ret = realloc(ret, (rc + 1) * sizeof(char*));
			assert(ret);
			ret[rc++] = strdup(line_translated);
			j = 0;
			continue;
		}
		if(line[i] == '\\')
		{
			if(line[i + 1])
			{
				if(line[i + 1] == 'n')
					line_translated[j++] = '\n';
				else
					line_translated[j++] = line[i + 1];
					
				i++;
			}
		}
		else
			line_translated[j++] = line[i];
	}
	line_translated[j++] = 0;
	ret = realloc(ret, (rc + 2) * sizeof(char*));
	ret[rc++] = strdup(line_translated);
	ret[rc++] = NULL;
	return(ret);
}

char * encode_line(char ** strtab, int n)
{
	int i, sl = 0;
	char * ret = NULL;
	
	for(i = 0; i < n; i++)
	{
		char * str = strtab[i];
		char * tstr = (char*) malloc(strlen(strtab[i]) * 2 + 2); /* the worst case */
		int j, k = 0;
		
		assert(tstr);
		tstr[k++] = '|';
		for(j = 0; str[j]; j++)
		{
			if(str[j] == '|')
			{
				tstr[k++] = '\\';
				tstr[k++] = '|';
			}
			else if(str[j] == '\n')
			{
				tstr[k++] = '\\';
				tstr[k++] = 'n';
			}
			else if(str[j] == '\\')
			{
				tstr[k++] = '\\';
				tstr[k++] = '\\';
			}
			else
				tstr[k++] = str[j];
		}
		tstr[k++] = 0;
		ret = realloc(ret, (sl + k) + 1);
		memcpy(ret + sl, tstr, k + 1);
		sl += k - 1;
		
		free(tstr);
	}
	
	return(ret);
}

static void free_strtab(char ** tab)
{
	int i = 0;
	while(tab[i])
		free(tab[i++]);
	
	free(tab);
}

static int count_strtab(char ** tab)
{
	int i = 0;
	while(tab[i]) i++;
	return(i);
}

void save_args(char * name, char ** argv, int argc)
{
	FILE *ifh;
	FILE *ofh;

	struct passwd * pw = getpwuid(getuid());
	char * path1;
	char * path2;
	int stored = 0;
	
	if(!pw)
	{
		fprintf(stderr, "Unable to save operation mode, unable to get information on user with UID=%d\n", getuid());
		exit(1);
	}

	path1 = (char*) malloc(strlen(pw->pw_dir) + 30);
	path2 = (char*) malloc(strlen(pw->pw_dir) + 30);
	sprintf(path1, "%s/.rate.modes", pw->pw_dir);
	sprintf(path2, "%s/.rate.modes.new", pw->pw_dir);

	if(!(ofh = fopen(path2, "w")))
	{
		perror("fopen");
		fprintf(stderr, "Unable to save operation mode, unable to open file '%s' for writing.\n");
		exit(1);
	}

	if((ifh = fopen(path1, "r")))
	{
		char line[MAX_LINE_SIZE];
		char ** translated;
		
		while(fgets(line, MAX_LINE_SIZE, ifh))
		{
			int i, l = strlen(line);

			translated = decode_line(line);
			
			if(strcmp(translated[0], name))
				fprintf(ofh, "%s\n", line);
			else
			{
				fprintf(ofh, "%s%s\n", name, encode_line(argv, argc));
				stored = 1;
			}
			
			free_strtab(translated);
		}
		
		fclose(ifh);
	}
	
	if(!stored)
		fprintf(ofh, "%s%s\n", name, encode_line(argv, argc));
	
	fclose(ofh);
	if(rename(path2, path1))
	{
		perror("rename");
		fprintf(stderr, "Unable to save operation mode, unable to rename file '%s' to '%s'\n", path2, path1);
		exit(1);
	}
	free(path1);
	free(path2);
}

int recall_args(char * name, char *** argv, int *argc)
{
	FILE *ifh;
	struct passwd * pw = getpwuid(getuid());
	char * path1;
	
	if(!pw)
	{
		fprintf(stderr, "Unable to load operation mode, unable to get information on user with UID=%d\n", getuid());
		exit(1);
	}

	path1 = (char*) malloc(strlen(pw->pw_dir) + 30);
	sprintf(path1, "%s/.rate.modes", pw->pw_dir);

	if((ifh = fopen(path1, "r")))
	{
		char line[MAX_LINE_SIZE];
		char ** translated;
		
		while(fgets(line, MAX_LINE_SIZE, ifh))
		{
			int i, l = strlen(line);

			translated = decode_line(line);
			if(strcmp(translated[0], name) == 0)
			{
				fclose(ifh);
				*argv = translated + 1;
				*argc = count_strtab(translated) - 1;
				free(path1);
				return(1);
			}
			free_strtab(translated);
		}
		fclose(ifh);
		fprintf(stderr, "No such operation mode: '%s' was stored.\n", name);
		exit(1);
	}
	else
	{
		perror("fopen");
		fprintf(stderr, "Unable to load operation mode, unable to open file '%s' for reading.\n", path1);
		exit(1);
	}
}
