/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifdef STREAM_ANALYZER
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/timeb.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <unistd.h>
#include "lib.h"
#include "streams.h"



static sta_stream * streamlist = NULL;

static sta_stream * streamhash[STREAMHASH_SIZE];
static sta_host * hosthash[HOSTHASH_SIZE];
static sta_stream * protolist[256];
static sta_class * classlist = NULL;


static unsigned int stream_counter = 0;
static unsigned int stale_stream_counter = 0;
static unsigned int host_counter = 0;

static unsigned int T_expiry_icmp = 300,
					T_expiry_udp = 1200,
					T_expiry_tcp = 3600,
					T_expiry_tcp_closed = 600,
					T_expiry_misc = 600;

static unsigned int T_stale_icmp = 5,
					T_stale_udp = 15,
					T_stale_tcp = 30,
					T_stale_misc = 30;
				
void stastream_addclass(u_int32_t ip, u_int32_t mask)
{
        sta_class * c = (sta_class*) sta_malloc(sizeof(sta_class));
        c->next = classlist;
        c->ip = ip & mask;
        c->mask = mask;
        classlist = c;
}

int stastream_isours(u_int32_t ip)
{
	sta_class * c = classlist;
	for(; c; c = c->next)
		if(c->ip == (ip & c->mask)) return(1);
	return(0);
}

sta_host ** stastream_gethosthash(void)
{
	return(hosthash);
}

sta_stream * stastream_getprotolist(int n)
{
	return(protolist[n]);
}

sta_stream ** stastream_getstreamhash(void)
{
	return(streamhash);
}

sta_stream * stastream_getstreamlist(void)
{
	return(streamlist);
}
	
					

static inline int stastream_hash_stream(u_int32_t e1, u_int32_t e2, u_int16_t p1, u_int16_t p2)
{
	return((e1 + e2 + p1 + p2) % STREAMHASH_SIZE);
}

static inline int stastream_hash_host(u_int32_t h)
{
	return(h % HOSTHASH_SIZE);
}

unsigned int stastream_streamcount(void)
{
	return(stream_counter);
}
unsigned int stastream_stalestreamcount(void)
{
	return(stale_stream_counter);
}

unsigned int stastream_hostcount(void)
{
	return(host_counter);
}

void stastream_hashinit(void)
{
	memset(streamhash, 0, sizeof(streamhash));
	memset(hosthash, 0, sizeof(hosthash));
	memset(protolist, 0, sizeof(protolist));
}

sta_host * stastream_findhost(u_int32_t ip)
{
	sta_host * host;
	int h, i;
	for(host = hosthash[h = stastream_hash_host(ip)]; host; host = host->hh_next)
		if(host->ip == ip) return(host);
	
	host = (sta_host*) sta_malloc(sizeof(sta_host));
	
	host->ip = ip;
	host->streamlist = NULL;
	host->streamcount = 0;
	memset(host->bytes, 0, 4 * sizeof(unsigned long long));
	memset(host->packets, 0, 4 * sizeof(unsigned long int));
	host->hh_next = hosthash[h];
	hosthash[h] = host;
	host_counter ++;
	return(host);
}

int counter = 0;

void stastream_classify(const unsigned char * ippkt, int len, int total_len, struct timeval * now)
{

	u_int32_t e1, e2;
	u_int16_t p1 = 0, p2 = 0;
	u_int16_t p = SPROP_PROTO_OTHER;
	u_int8_t ttl;
	int proto, closed = 0, ignore = 0;
	
	const unsigned char * protohdr;
	const unsigned char * content;
	
	sta_stream * s;
	int i, h, endpoint = 0, offset;

	if(len < sizeof(struct ip)) return;
	if(((struct ip*) ippkt)->ip_v != 4) return;
	
	e1 = ((struct ip*) ippkt)->ip_src.s_addr;
	e2 = ((struct ip*) ippkt)->ip_dst.s_addr;
	ttl = ((struct ip*) ippkt)->ip_ttl;
	proto = ((struct ip*) ippkt)->ip_p;
	
	offset = ((((struct ip*) ippkt)->ip_hl & 0xf) << 2);
	
	if(offset > len) return;
	
	protohdr = ippkt + offset;
	len -= offset;

	
	switch(proto)
	{
		case IPPROTO_TCP:
			if(len < sizeof(struct tcphdr)) return;
			p = SPROP_PROTO_TCP;
			p1 = ((struct tcphdr*) protohdr)->th_sport;
			p2 = ((struct tcphdr*) protohdr)->th_dport;
			offset = ((struct tcphdr*) protohdr)->th_off << 2;
			if(offset > len) return;
			
			if(((struct tcphdr*) protohdr)->th_flags & (TH_RST | TH_FIN))
			   	closed = ignore = 1;
			   
			if(((struct tcphdr*) protohdr)->th_flags & TH_SYN)
				ignore = 1;
		
			len -= offset;
			content = protohdr + offset;
			break;
			
		case IPPROTO_UDP:
			if(len < sizeof(struct udphdr)) return;
			p = SPROP_PROTO_UDP;
			p1 = ((struct udphdr*) protohdr)->uh_sport;
			p2 = ((struct udphdr*) protohdr)->uh_dport;
			content = protohdr + sizeof(struct udphdr);
			len -= sizeof(struct udphdr);
			break;		
		case IPPROTO_ICMP:
			p = SPROP_PROTO_ICMP;
		default:
			content = protohdr;
			break;
	}

	
	
	for(s = streamhash[h = stastream_hash_stream(e1, e2, p1, p2)]; s; s = s->sh_next)
	{
		debug(50, "Stream lookup @ hash %d: %x\n", h, s);
		if((s->properties & SPROP_PROTO_MASK) != p)
		{
			debug(50, "(Wrong protocol: %x (%x) and %x)\n", s->properties, s->properties & SPROP_PROTO_MASK, p);
		 	continue;
		}

		if(s->endpoints[0].host->ip == e1)
			if(s->endpoints[1].host->ip == e2)
				if(s->endpoints[0].port == p1)
					if(s->endpoints[1].port == p2)
						break;

		if(s->endpoints[1].host->ip == e1)
			if(s->endpoints[0].host->ip == e2)
				if(s->endpoints[1].port == p1)
					if(s->endpoints[0].port == p2)
					{
						u_int32_t temp1;
						u_int16_t temp2;
						
						temp1 = e1;
						e1 = e2;
						e2 = temp1;
						temp2 = p1;
						p1 = p2;
						p2 = temp2;
						
						endpoint = 1;
						
						if(s->properties | SPROP_UNIDIRECTIONAL)
						{
							s->properties &= ~SPROP_UNIDIRECTIONAL;
							memcpy(&s->endpoints[1].firstdata, now, sizeof(struct timeval));
						}
						break;
					}
	}


	
	if(!s)
	{
		sta_host * host1, * host2;
		
		s = (sta_stream*) sta_malloc(sizeof(sta_stream));
		
		host1 = stastream_findhost(e1);
		host2 = stastream_findhost(e2);
	
		s->properties = SPROP_UNIDIRECTIONAL | SPROP_OPENED | p;
		s->endpoints[0].host = host1;
		s->endpoints[0].port = p1;
		s->endpoints[0].lastbytes = 0;
		s->endpoints[0].lastpackets = 0;
		s->endpoints[0].bytes = 0;
		s->endpoints[0].packets = 0;
		s->endpoints[0].dpackets = 0;
		s->endpoints[1].host = host2;
		s->endpoints[1].port = p2;
		s->endpoints[1].lastbytes = 0;
		s->endpoints[1].lastpackets = 0;
		s->endpoints[1].bytes = 0;
		s->endpoints[1].packets = 0;
		s->endpoints[1].dpackets = 0;
	
		s->stream_protocol = 0;
		s->protocol_data = NULL;
		s->protocol_level = 0;
		
		memcpy(&s->endpoints[0].firstdata, now, sizeof(struct timeval));
		memset(&s->endpoints[1].firstdata, 0, sizeof(struct timeval));
		
		s->sh_next = streamhash[h];
		if(streamhash[h])
			streamhash[h]->sh_unlinkptr = &s->sh_next;
		streamhash[h] = s;
		s->sh_unlinkptr = &streamhash[h];
		
		s->stal_next = streamlist;
		if(streamlist)
			streamlist->stal_unlinkptr = &s->stal_next;
		streamlist = s;
		s->stal_unlinkptr = &streamlist;

		s->proto_next = NULL;
		s->proto_unlinkptr = NULL;

		s->endpoints[0].sl_next = host1->streamlist;
		if(host1->streamlist)
		{
			if(host1->streamlist->endpoints[0].host == host1)
				host1->streamlist->endpoints[0].sl_unlinkptr = &s->endpoints[0].sl_next;
			else
				host1->streamlist->endpoints[1].sl_unlinkptr = &s->endpoints[0].sl_next;
		}
		host1->streamlist = s;
		s->endpoints[0].sl_unlinkptr = &host1->streamlist;

		s->endpoints[1].sl_next = host2->streamlist;
		if(host2->streamlist)
		{
			if(host2->streamlist->endpoints[0].host == host2)
				host2->streamlist->endpoints[0].sl_unlinkptr = &s->endpoints[1].sl_next;
			else
				host2->streamlist->endpoints[1].sl_unlinkptr = &s->endpoints[1].sl_next;
		}
		host2->streamlist = s;
		s->endpoints[1].sl_unlinkptr = &host2->streamlist;

		host1->streamcount ++;
		host2->streamcount ++;
		stream_counter++;
		
		debug(9, "New stream, overall streams: %d\n", stream_counter);
	}


	s->endpoints[endpoint].host->bytes[LSENT] += total_len;
	s->endpoints[endpoint].host->packets[LSENT] ++;
	s->endpoints[1 - endpoint].host->bytes[LRECV] += total_len;
	s->endpoints[1 - endpoint].host->packets[LRECV] ++;

	s->endpoints[endpoint].lastbytes += total_len;
	s->endpoints[endpoint].lastpackets ++;
	memcpy(&s->endpoints[endpoint].lastdata, now, sizeof(struct timeval));
	
	debug(49, "Starting AN\n");

	
	/* content analysis */
	if(!ignore)
	{
		unsigned short counters[256];
		int eightbit = 0, i, l, compr = 1;
		int crv, bv;
		unsigned char proto;
	
		s->endpoints[endpoint].dpackets ++;
	
		memset(counters, 0, sizeof(unsigned short) * 256);
		l = len;
		if(l > SNAPLEN) l = SNAPLEN;
		
		debug(49, "Will inspect %d bytes:", l);
		
		crv = l / COMPRESSION_RVALUE;
		bv = l / BINARY_RVALUE;
		if(bv < BINARY_MINIMAL) bv = BINARY_MINIMAL;
		
		for(i = 0; i < l; i++)
		{
			unsigned char c = content[i];
			debug(49, "%02x", c);
			counters[c] ++;
			if(counters[c] > crv)
			{
				compr = 0;
				debug(49, "(occurs %d times, no compr)", counters[c]);
			}
				
			if(c & 0x80)
			{
				debug(49, "data: %c (%02x) is binary\n", c, c);
				eightbit++;
				if(eightbit > bv && compr == 0) break;
			}
		}
		debug(49, "\n");
			
			
		if(compr)
			s->properties |= SPROP_COMPRESSION_COMPRESSED;
		else
			s->properties &= ~SPROP_COMPRESSION_COMPRESSED;
		
		if(eightbit > bv)
			s->properties |= SPROP_BINARY;
		else
			s->properties &= ~SPROP_BINARY;	
			
		if((proto = sta_process(s, content, len, endpoint)))
		{
			s->stream_protocol = proto;

			s->proto_next = protolist[proto];
			if(protolist[proto])
				protolist[proto]->proto_unlinkptr = &s->proto_next;
			protolist[proto] = s;
			s->proto_unlinkptr = &protolist[proto];
		}
	}
	
	if(closed)
		s->properties |= SPROP_CLOSED;

	debug(49, "Writing DATA\n");


	debug(10, "Data from %x Stream T=%d: %08x.%d (%d) => %08x.%d (%d) %s%s%s%s\n",
					e1, s->stream_protocol,
					s->endpoints[0].host->ip,
					ntohs(s->endpoints[0].port),
					s->endpoints[0].lastbytes,
					s->endpoints[1].host->ip,
					ntohs(s->endpoints[1].port),
					s->endpoints[1].lastbytes,
					(s->properties & SPROP_UNIDIRECTIONAL) ? "(UNIDIRECTIONAL)" : "",
					(s->properties & SPROP_BINARY) ? "(BINARY)" : "",
					(s->properties & SPROP_COMPRESSION_COMPRESSED) ? "(COMPRESSED)" : "",
					(s->properties & SPROP_CLOSED) ? "(CLOSED)" : "");

	
}

void stastream_expire(struct timeb * now)
{
	sta_stream * s, * next;
	int ssc = 0, i;

	debug(41, "stastream_expire(): Entering\n");
	
	for(s = streamlist; s; s = next)
	{
		int et, st;
		next = s->stal_next;
		
		switch(s->properties & SPROP_PROTO_MASK)
		{
			case SPROP_PROTO_TCP:
				if(s->properties & SPROP_CLOSED)
					et = T_expiry_tcp_closed;
				else
					et = T_expiry_tcp;
				st = T_stale_tcp;
				break;
			case SPROP_PROTO_UDP: et = T_expiry_udp; st = T_stale_udp; break;
			case SPROP_PROTO_ICMP: et = T_expiry_icmp; st = T_stale_icmp; break;
			default: et = T_expiry_misc; st = T_stale_misc; break;
		}
		if((now->time - s->endpoints[0].lastdata.tv_sec) > et &&
		   (now->time - s->endpoints[1].lastdata.tv_sec) > et)
		{
			stream_counter--;
			
			debug(49, "Unlinking stream...\n");
			
			*s->sh_unlinkptr = s->sh_next;
			*s->stal_unlinkptr = s->stal_next;
			if(s->proto_unlinkptr) *s->proto_unlinkptr = s->proto_next;
			*s->endpoints[0].sl_unlinkptr = s->endpoints[0].sl_next;
			*s->endpoints[1].sl_unlinkptr = s->endpoints[1].sl_next;


			if(s->sh_next) s->sh_next->sh_unlinkptr = s->sh_unlinkptr;
			if(s->stal_next) s->stal_next->stal_unlinkptr = s->stal_unlinkptr;
			if(s->proto_next) s->proto_next->proto_unlinkptr = s->proto_unlinkptr;
			if(s->endpoints[0].sl_next)
			{
				if(s->endpoints[0].host == s->endpoints[0].sl_next->endpoints[0].host)
					s->endpoints[0].sl_next->endpoints[0].sl_unlinkptr = s->endpoints[0].sl_unlinkptr;
				else
					s->endpoints[0].sl_next->endpoints[1].sl_unlinkptr = s->endpoints[0].sl_unlinkptr;
			}
			if(s->endpoints[1].sl_next)
			{
				if(s->endpoints[1].host == s->endpoints[1].sl_next->endpoints[0].host)
					s->endpoints[1].sl_next->endpoints[0].sl_unlinkptr = s->endpoints[1].sl_unlinkptr;
				else
					s->endpoints[1].sl_next->endpoints[1].sl_unlinkptr = s->endpoints[1].sl_unlinkptr;
			}


			s->endpoints[0].host->streamcount--;
			s->endpoints[1].host->streamcount--;
			
			sta_destroyprotodata(s);
			sta_free(s);
			continue;
		}

		if((now->time - s->endpoints[0].lastdata.tv_sec) > st &&
		   (now->time - s->endpoints[1].lastdata.tv_sec) > st)
		{
			s->properties |= SPROP_STALE;
			ssc ++;
		}
		else
			s->properties &= ~SPROP_STALE;
		
		s->endpoints[0].bytes += s->endpoints[0].lastbytes;
		s->endpoints[0].packets += s->endpoints[0].lastpackets;

		s->endpoints[1].bytes += s->endpoints[0].lastbytes;
		s->endpoints[1].packets += s->endpoints[0].lastpackets;
		
	}

	for(i = 0; i<HOSTHASH_SIZE; i++)
	{
		sta_host * h = hosthash[i];
		sta_host * next;

		for(; h; h = h->hh_next)
		{
			h->bytes[SENT] += h->bytes[LSENT];
			h->packets[SENT] += h->packets[LSENT];
			h->bytes[RECV] += h->bytes[LRECV];
			h->packets[RECV] += h->packets[LRECV];
		}
	}


	stale_stream_counter = ssc;
	debug(41, "stastream_expire(): Leaving\n");

}

void stastream_clear(void)
{
	int i;
	sta_stream * s;
	for(s = streamlist; s; s = s->stal_next)
	{
		s->endpoints[0].lastbytes = 0;
		s->endpoints[0].lastpackets = 0;
		s->endpoints[1].lastbytes = 0;
		s->endpoints[1].lastpackets = 0;
	}
	for(i = 0; i<HOSTHASH_SIZE; i++)
	{
		sta_host * h = hosthash[i];
		sta_host * next;

		for(; h; h = h->hh_next)
		{
			h->bytes[LRECV] = h->bytes[LSENT] = 0;
			h->packets[LRECV] = h->packets[LSENT] = 0;
		}
	}
}
void stastream_cleanstreams(struct timeb *now, int count)
{
	int *leasttimes = (void*) sta_malloc(sizeof(int) * count);
	sta_stream ** leasttable = (void*) sta_malloc(sizeof(sta_stream*) * count);
	sta_stream * s, * next;
	int i;

	memset(leasttable, 0, sizeof(sta_stream*) * count);
	memset(leasttimes, 0, sizeof(int) * count);
	
	for(s = streamlist; s; s = next)
	{
		int dist, dist2, i;
		next = s->stal_next;

		dist = now->time - s->endpoints[0].lastdata.tv_sec;
		dist2 = now->time - s->endpoints[1].lastdata.tv_sec;
		
		if(dist2 < dist)
			dist = dist2;

		if(dist < leasttimes[count - 1]) continue;

		for(i = 0; i < count; i++)
			if(dist > leasttimes[i])
			{
				if(i != count)
				{
					memmove(leasttimes + i + 1, leasttimes + i, sizeof(int) * (count - i - 1));
					memmove(leasttable + i + 1, leasttable + i, sizeof(sta_stream*) * (count - i - 1));
				}
				leasttimes[i] = dist;
				leasttable[i] = s;
				break;
			}
	}
	
	for(i = 0; i<count; i++)
	{

		s = leasttable[i];
		if(!s) continue;
	
		stream_counter--;
		debug(49, "Unlinking stream (cleaning)...\n");
			
		*s->sh_unlinkptr = s->sh_next;
		*s->stal_unlinkptr = s->stal_next;
		if(s->proto_unlinkptr) *s->proto_unlinkptr = s->proto_next;

		*s->endpoints[0].sl_unlinkptr = s->endpoints[0].sl_next;
		*s->endpoints[1].sl_unlinkptr = s->endpoints[1].sl_next;

		if(s->sh_next) s->sh_next->sh_unlinkptr = s->sh_unlinkptr;
		if(s->stal_next) s->stal_next->stal_unlinkptr = s->stal_unlinkptr;
		if(s->proto_next) s->proto_next->proto_unlinkptr = s->proto_unlinkptr;
		
		if(s->endpoints[0].sl_next)
		{
			if(s->endpoints[0].host == s->endpoints[0].sl_next->endpoints[0].host)
				s->endpoints[0].sl_next->endpoints[0].sl_unlinkptr = s->endpoints[0].sl_unlinkptr;
			else
				s->endpoints[0].sl_next->endpoints[1].sl_unlinkptr = s->endpoints[0].sl_unlinkptr;
		}
		if(s->endpoints[1].sl_next)
		{
			if(s->endpoints[1].host == s->endpoints[1].sl_next->endpoints[0].host)
				s->endpoints[1].sl_next->endpoints[0].sl_unlinkptr = s->endpoints[1].sl_unlinkptr;
			else
				s->endpoints[1].sl_next->endpoints[1].sl_unlinkptr = s->endpoints[1].sl_unlinkptr;
		}

		s->endpoints[0].host->streamcount--;
		s->endpoints[1].host->streamcount--;
		
		sta_destroyprotodata(s);
		sta_free(s);
	}

	for(i = 0; i<HOSTHASH_SIZE; i++)
	{
		sta_host * h = hosthash[i];
		sta_host * next;
		for(; h; h = h->hh_next)
		{
			if((next = h->hh_next))
			{
				if(!(next->streamcount || stastream_isours(next->ip)))
				{
					h->hh_next = next->hh_next;
					free(next);
					host_counter --;
				}
			}
		}
	}
	free(leasttimes);
	free(leasttable);
}
#endif
