/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#ifdef DEBUG
#include <stdio.h>
#define debug(level, format, arg...) if(level <= DEBUG) fprintf(stderr, format, ## arg)
#else
#define debug(level, format, arg...)
#endif

#define STREAMHASH_SIZE 13041
#define HOSTHASH_SIZE 3943

#define SNAPLEN 512
#define COMPRESSION_RVALUE 32
#define BINARY_RVALUE 4
#define BINARY_MINIMAL 5

struct class
{
        struct class * next;
        u_int32_t ip;
        u_int32_t mask;
};

typedef struct class sta_class;

#define SENT 0
#define RECV 1
#define LSENT 2
#define LRECV 3

struct host
{
	struct host * hh_next;
	
	u_int32_t ip;

	unsigned long long bytes[4];
	unsigned long int  packets[4];
	
	int streamcount;
	struct stream * streamlist;
};

typedef struct host sta_host;

struct stream
{
	struct stream * sh_next;
	struct stream * stal_next;
	struct stream * proto_next;
	
	struct stream ** sh_unlinkptr;
	struct stream ** stal_unlinkptr;
	struct stream ** proto_unlinkptr;
	
	struct endpoint
	{
		struct host * host;

		struct stream * sl_next;
		struct stream ** sl_unlinkptr;

		u_int16_t port;
		
		unsigned int lastbytes;
		unsigned int lastpackets;

		unsigned int bytes;
		unsigned int packets;

		unsigned int dpackets;

		u_int8_t ttl;		
		
		struct timeval firstdata;
		struct timeval lastdata;
	} endpoints[2];

	u_int16_t properties;
		/* stream properties, see SPROP_ */

	u_int16_t classifier_count;

	u_int8_t stream_protocol;
	u_int8_t protocol_level;
	void * protocol_data;

	u_int8_t data_type;
};

typedef struct stream sta_stream;

#define STREAM_UNKNOWN 0

#define STREAM_GENERIC_TRANSFER 1
#define STREAM_GENERIC_INTERACTIVE 2
#define STREAM_GENERIC_CONVERSATION 3

#define STREAM_ICMP_PROBE 4
#define STREAM_ICMP_MISC 5

#define STREAM_HTTP	6
#define STREAM_FTP_CONTROL 7
#define STREAM_FTP_DATA 8
#define STREAM_SMTP 9
#define STREAM_POP3 10
#define STREAM_DOMAIN 11
#define STREAM_SSH 12

#define STREAM_TELNET 13
#define STREAM_TFTP 14
/*
    rate  --  statistic traffic analyzer
    Copyright (C) 2003 Mateusz 'mteg' Golicz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define STREAM_AUTH 15

#define STREAM_MAX 16

#define LEVEL_GUESS 0xff

/* sizeof(struct stream) == 66 bytes */

#define SPROP_PROTO_OTHER	0x0000
#define SPROP_PROTO_TCP		0x4000
#define SPROP_PROTO_UDP		0x8000
#define	SPROP_PROTO_ICMP	0xc000

#define SPROP_PROTO_MASK	0xc000

#define SPROP_COMPRESSION_UNCOMPRESSED	0x0000
#define SPROP_COMPRESSION_COMPRESSED	0x1000
#define SPROP_COMPRESSION_MIXED			0x2000
#define SPROP_ASCII		0x0000
#define SPROP_BINARY	0x0400
#define SPROP_CIPHER_UNCIPHERED	0x0000
#define SPROP_CIPHER_CIPHERED	0x0001
#define SPROP_CIPHER_MIXED		0x0002

#define SPROP_UNIDIRECTIONAL	0x0004
#define SPROP_BIDIRECTIONAL		0x0000

#define SPROP_CLOSED			0x0040
#define SPROP_OPENED			0x0000
#define SPROP_DEAD				0x0010
#define SPROP_STALE				0x0020

unsigned int stastream_streamcount(void);
unsigned int stastream_hostcount(void);
void stastream_hashinit(void);
sta_host * stastream_findhost(u_int32_t ip);
void stastream_classify(const unsigned char * ippkt, int len, int, struct timeval * now);
void stastream_expire(struct timeb * now);
unsigned int stastream_stalestreamcount(void);
void stastream_addclass(u_int32_t ip, u_int32_t mask);
sta_host ** stastream_gethosthash(void);
sta_stream ** stastream_getstreamhash(void);
void stastream_cleanstreams(struct timeb *now, int count);
void stastream_clear(void);
sta_stream * stastream_getstreamlist(void);
sta_stream * stastream_getprotolist(int n);
